@@ -6,7 +6,6 @@
 # either the dual license version in 2003, or any later version.
 
 
-require 'uri'
 require 'stringio'
 require 'digest/sha1'
 
@@ -65,7 +64,7 @@ require 'httpclient/cookie'
 #
 # 3. You can pass :follow_redirect option to follow redirect response in get.
 #
-#     puts clnt.get('http://dev.ctor.org/', :foolow_redirect => true)
+#     puts clnt.get('http://dev.ctor.org/', :follow_redirect => true)
 #
 # 4. Get content as chunks of String.  It yields chunks of String.
 #
@@ -198,7 +197,7 @@ require 'httpclient/cookie'
 #
 # Pass a Hash or an Array for header argument.
 #
-#     header = { 'Accept' => '*/*' }
+#     header = { 'Accept' => 'text/html' }
 #     clnt.get(uri, query, header)
 #
 #     header = [['Accept', 'image/jpeg'], ['Accept', 'image/png']]
@@ -230,7 +229,7 @@ require 'httpclient/cookie'
 #   ruby -rhttpclient -e 'p HTTPClient.head(ARGV.shift).header["last-modified"]' http://dev.ctor.org/
 #
 class HTTPClient
-  RUBY_VERSION_STRING = "ruby #{RUBY_VERSION} (#{RUBY_RELEASE_DATE}) [#{RUBY_PLATFORM}]"
+  RUBY_VERSION_STRING = "ruby #{RUBY_VERSION} (#{RUBY_RELEASE_DATE})"
   LIB_NAME = "(#{VERSION}, #{RUBY_VERSION_STRING})"
 
   include Util
@@ -358,6 +357,9 @@ class HTTPClient
   # Default header for PROPFIND request.
   PROPFIND_DEFAULT_EXTHEADER = { 'Depth' => '0' }
 
+  # Default User-Agent header
+  DEFAULT_AGENT_NAME = 'HTTPClient/1.0'
+
   # Creates a HTTPClient instance which manages sessions, cookies, etc.
   #
   # HTTPClient.new takes 3 optional arguments for proxy url string,
@@ -378,6 +380,7 @@ class HTTPClient
     proxy, agent_name, from = keyword_argument(args, :proxy, :agent_name, :from)
     @proxy = nil        # assigned later.
     @no_proxy = nil
+    @no_proxy_regexps = []
     @www_auth = WWWAuth.new
     @proxy_auth = ProxyAuth.new
     @request_filter = [@proxy_auth, @www_auth]
@@ -385,7 +388,7 @@ class HTTPClient
     @redirect_uri_callback = method(:default_redirect_uri_callback)
     @test_loopback_response = []
     @session_manager = SessionManager.new(self)
-    @session_manager.agent_name = agent_name
+    @session_manager.agent_name = agent_name || DEFAULT_AGENT_NAME
     @session_manager.from = from
     @session_manager.ssl_config = @ssl_config = SSLConfig.new(self)
     @cookie_manager = WebAgent::CookieManager.new
@@ -478,6 +481,17 @@ class HTTPClient
   # Calling this method resets all existing sessions.
   def no_proxy=(no_proxy)
     @no_proxy = no_proxy
+    @no_proxy_regexps.clear
+    if @no_proxy
+      @no_proxy.scan(/([^:,]+)(?::(\d+))?/) do |host, port|
+        if host[0] == ?.
+          regexp = /#{Regexp.quote(host)}\z/i
+        else
+          regexp = /(\A|\.)#{Regexp.quote(host)}\z/i
+        end
+        @no_proxy_regexps << [regexp, port]
+      end
+    end
     reset_all
   end
 
@@ -562,7 +576,7 @@ class HTTPClient
   #         Give an array to pass multiple value like
   #         [["a", "b"], ["a", "c"]] => 'http://host/part?a=b&a=c'.
   # header:: a Hash or an Array of extra headers.  e.g.
-  #          { 'Accept' => '*/*' } or
+  #          { 'Accept' => 'text/html' } or
   #          [['Accept', 'image/jpeg'], ['Accept', 'image/png']].
   # &block:: Give a block to get chunked message-body of response like
   #          get_content(uri) { |chunked_body| ... }.
@@ -577,7 +591,7 @@ class HTTPClient
   # follow HTTP redirect by yourself if you need.
   def get_content(uri, *args, &block)
     query, header = keyword_argument(args, :query, :header)
-    follow_redirect(:get, uri, query, nil, header || {}, &block).content
+    success_content(follow_redirect(:get, uri, query, nil, header || {}, &block))
   end
 
   # Posts a content.
@@ -597,7 +611,7 @@ class HTTPClient
   #           { 'Content-Type' => 'video/mp4', :content => File.new('video.mp4') }]
   #          => <Two parts with custom Content-Type header>
   # header:: a Hash or an Array of extra headers. e.g.
-  #            { 'Accept' => '*/*' }
+  #            { 'Accept' => 'text/html' }
   #          or
   #            [['Accept', 'image/jpeg'], ['Accept', 'image/png']].
   # &block:: Give a block to get chunked message-body of response like
@@ -607,12 +621,14 @@ class HTTPClient
   # post_content follows HTTP redirect status (see HTTP::Status.redirect?)
   # internally and try to post the content to redirected URL.  See
   # redirect_uri_callback= how HTTP redirection is handled.
+  # Bear in mind that you should not depend on post_content because it sends
+  # the same POST method to the new location which is prohibited in HTTP spec.
   #
   # If you need to get full HTTP response including HTTP status and headers,
   # use post method.
   def post_content(uri, *args, &block)
     body, header = keyword_argument(args, :body, :header)
-    follow_redirect(:post, uri, nil, body, header || {}, &block).content
+    success_content(follow_redirect(:post, uri, nil, body, header || {}, &block))
   end
 
   # A method for redirect uri callback.  How to use:
@@ -621,11 +637,11 @@ class HTTPClient
   #   Location: ../foo/
   # in HTTP header. (raises BadResponseError instead)
   def strict_redirect_uri_callback(uri, res)
-    newuri = URI.parse(res.header['location'][0])
+    newuri = urify(res.header['location'][0])
     if https?(uri) && !https?(newuri)
       raise BadResponseError.new("redirecting to non-https resource")
     end
-    unless newuri.is_a?(URI::HTTP)
+    if !http?(newuri) && !https?(newuri)
       raise BadResponseError.new("unexpected location: #{newuri}", res)
     end
     puts "redirect to: #{newuri}" if $DEBUG
@@ -638,11 +654,11 @@ class HTTPClient
   #   Location: ../foo/
   # in HTTP header.
   def default_redirect_uri_callback(uri, res)
-    newuri = URI.parse(res.header['location'][0])
-    unless newuri.is_a?(URI::HTTP)
+    newuri = urify(res.header['location'][0])
+    if !http?(newuri) && !https?(newuri)
       newuri = uri + newuri
-      STDERR.puts("could be a relative URI in location header which is not recommended")
-      STDERR.puts("'The field value consists of a single absolute URI' in HTTP spec")
+      warn("could be a relative URI in location header which is not recommended")
+      warn("'The field value consists of a single absolute URI' in HTTP spec")
     end
     if https?(uri) && !https?(newuri)
       raise BadResponseError.new("redirecting to non-https resource")
@@ -653,7 +669,7 @@ class HTTPClient
 
   # Sends HEAD request to the specified URL.  See request for arguments.
   def head(uri, *args)
-    request(:head, uri, argument_to_hash(args, :query, :header))
+    request(:head, uri, argument_to_hash(args, :query, :header, :follow_redirect))
   end
 
   # Sends GET request to the specified URL.  See request for arguments.
@@ -662,6 +678,8 @@ class HTTPClient
   end
 
   # Sends POST request to the specified URL.  See request for arguments.
+  # You should not depend on :follow_redirect => true for POST method.  It
+  # sends the same POST method to the new location which is prohibited in HTTP spec.
   def post(uri, *args, &block)
     request(:post, uri, argument_to_hash(args, :body, :header, :follow_redirect), &block)
   end
@@ -673,7 +691,7 @@ class HTTPClient
 
   # Sends DELETE request to the specified URL.  See request for arguments.
   def delete(uri, *args, &block)
-    request(:delete, uri, argument_to_hash(args, :header), &block)
+    request(:delete, uri, argument_to_hash(args, :body, :header), &block)
   end
 
   # Sends OPTIONS request to the specified URL.  See request for arguments.
@@ -693,7 +711,7 @@ class HTTPClient
   
   # Sends TRACE request to the specified URL.  See request for arguments.
   def trace(uri, *args, &block)
-    request('TRACE', uri, argument_to_hash(args, :query, :body, :header), &block)
+    request('TRACE', uri, argument_to_hash(args, :query, :header), &block)
   end
 
   # Sends a request to the specified URL.
@@ -721,7 +739,7 @@ class HTTPClient
   #          => <Two parts with custom Content-Type header>
   #        See HTTP::Message.file? for actual condition of 'a file'.
   # header:: a Hash or an Array of extra headers.  e.g.
-  #          { 'Accept' => '*/*' } or
+  #          { 'Accept' => 'text/html' } or
   #          [['Accept', 'image/jpeg'], ['Accept', 'image/png']].
   # &block:: Give a block to get chunked message-body of response like
   #          get(uri) { |chunked_body| ... }.
@@ -929,7 +947,7 @@ private
     uri = urify(uri)
     if block
       filtered_block = proc { |r, str|
-        block.call(str) if HTTP::Status.successful?(r.status)
+        block.call(str) if r.ok?
       }
     end
     if HTTP::Message.file?(body)
@@ -939,18 +957,25 @@ private
     while retry_number < @follow_redirect_count
       body.pos = pos if pos
       res = do_request(method, uri, query, body, header, &filtered_block)
-      if HTTP::Status.successful?(res.status)
-        return res
-      elsif HTTP::Status.redirect?(res.status)
+      if res.redirect?
+        method = :get if res.see_other? # See RFC2616 10.3.4
         uri = urify(@redirect_uri_callback.call(uri, res))
         retry_number += 1
       else
-        raise BadResponseError.new("unexpected response: #{res.header.inspect}", res)
+        return res
       end
     end
     raise BadResponseError.new("retry count exceeded", res)
   end
 
+  def success_content(res)
+    if res.ok?
+      return res.content
+    else
+      raise BadResponseError.new("unexpected response: #{res.header.inspect}", res)
+    end
+  end
+
   def protect_keep_alive_disconnected
     begin
       yield
@@ -984,7 +1009,7 @@ private
             header = override_header(header, 'Content-Type', content_type)
           end
         end
-      elsif method == 'POST'
+      else
         if file_in_form_data?(body)
           boundary = create_boundary
           content_type = "multipart/form-data; boundary=#{boundary}"
@@ -996,7 +1021,7 @@ private
     end
     req = HTTP::Message.new_request(method, uri, query, body, boundary)
     header.each do |key, value|
-      req.header.add(key, value)
+      req.header.add(key.to_s, value)
     end
     if @cookie_manager && cookie = @cookie_manager.find(uri)
       req.header.add('Cookie', cookie)
@@ -1031,13 +1056,11 @@ private
     if !@proxy or NO_PROXY_HOSTS.include?(uri.host)
       return true
     end
-    unless @no_proxy
-      return false
-    end
-    @no_proxy.scan(/([^:,]+)(?::(\d+))?/) do |host, port|
-      if /(\A|\.)#{Regexp.quote(host)}\z/i =~ uri.host &&
-          (!port || uri.port == port.to_i)
-        return true
+    @no_proxy_regexps.each do |regexp, port|
+      if !port || uri.port == port.to_i
+        if regexp =~ uri.host
+          return true
+        end
       end
     end
     false
@@ -1051,11 +1074,11 @@ private
     end
     if str = @test_loopback_response.shift
       dump_dummy_request_response(req.http_body.dump, str) if @debug_dev
-      conn.push(HTTP::Message.new_response(str))
+      conn.push(HTTP::Message.new_response(str, req.header))
       return
     end
     content = block ? nil : ''
-    res = HTTP::Message.new_response(content)
+    res = HTTP::Message.new_response(content, req.header)
     @debug_dev << "= Request\n\n" if @debug_dev
     sess = @session_manager.query(req, proxy)
     res.peer_cert = sess.ssl_peer_cert
@@ -1087,11 +1110,11 @@ private
     end
     if str = @test_loopback_response.shift
       dump_dummy_request_response(req.http_body.dump, str) if @debug_dev
-      conn.push(HTTP::Message.new_response(StringIO.new(str)))
+      conn.push(HTTP::Message.new_response(StringIO.new(str), req.header))
       return
     end
     piper, pipew = IO.pipe
-    res = HTTP::Message.new_response(piper)
+    res = HTTP::Message.new_response(piper, req.header)
     @debug_dev << "= Request\n\n" if @debug_dev
     sess = @session_manager.query(req, proxy)
     res.peer_cert = sess.ssl_peer_cert
