@@ -4,10 +4,12 @@
 # This program is copyrighted free software by NAKAMURA, Hiroshi.  You can
 # redistribute it and/or modify it under the same terms of Ruby's license;
 # either the dual license version in 2003, or any later version.
-
-# httpclient/session.rb is based on http-access.rb in http-access/0.0.4.
-# Some part of code in http-access.rb was recycled in httpclient.rb.
-# Those part is copyrighted by Maehashi-san.
+#
+# httpclient/session.rb is based on http-access.rb in http-access/0.0.4.  Some
+# part of it is copyrighted by Maebashi-san who made and published
+# http-access/0.0.4.  http-access/0.0.4 did not include license notice but when
+# I asked Maebashi-san he agreed that I can redistribute it under the same terms
+# of Ruby.  Many thanks to Maebashi-san.
 
 
 require 'socket'
@@ -29,6 +31,7 @@ class HTTPClient
     attr_accessor :scheme
     # Host String.
     attr_accessor :host
+    alias hostname host
     # Port number.
     attr_accessor :port
 
@@ -36,7 +39,7 @@ class HTTPClient
     def initialize(uri = nil)
       if uri
         @scheme = uri.scheme
-        @host = uri.host
+        @host = uri.hostname
         @port = uri.port.to_i
       else
         @scheme = 'tcp'
@@ -316,6 +319,18 @@ class HTTPClient
       end
     end
 
+    def ssl_version
+      @ssl_socket.ssl_version if @ssl_socket.respond_to?(:ssl_version)
+    end
+
+    def ssl_cipher
+      @ssl_socket.cipher
+    end
+
+    def ssl_state
+      @ssl_socket.state
+    end
+
     def peer_cert
       @ssl_socket.peer_cert
     end
@@ -711,16 +726,21 @@ class HTTPClient
           req.http_version = $1
         end
       end
-      if @agent_name
+      if @agent_name && req.header.get('User-Agent').empty?
         req.header.set('User-Agent', "#{@agent_name} #{LIB_NAME}")
       end
-      if @from
+      if @from && req.header.get('From').empty?
         req.header.set('From', @from)
       end
+      if req.header.get('Accept').empty?
+        req.header.set('Accept', '*/*')
+      end
       if @transparent_gzip_decompression
         req.header.set('Accept-Encoding', 'gzip,deflate')
       end
-      req.header.set('Date', Time.now.httpdate)
+      if req.header.get('Date').empty?
+        req.header.set_date_header
+      end
     end
 
     # Connect to the server
@@ -732,11 +752,19 @@ class HTTPClient
           @socket = create_socket(site)
           if https?(@dest)
             if @socket.is_a?(LoopBackSocket)
-              connect_ssl_proxy(@socket, URI.parse(@dest.to_s)) if @proxy
+              connect_ssl_proxy(@socket, urify(@dest.to_s)) if @proxy
             else
               @socket = create_ssl_socket(@socket)
-              connect_ssl_proxy(@socket, URI.parse(@dest.to_s)) if @proxy
-              @socket.ssl_connect(@dest.host)
+              connect_ssl_proxy(@socket, urify(@dest.to_s)) if @proxy
+              begin
+                @socket.ssl_connect(@dest.host)
+              ensure
+                if $DEBUG
+                  warn("Protocol version: #{@socket.ssl_version}")
+                  warn("Cipher: #{@socket.ssl_cipher.inspect}")
+                  warn("State: #{@socket.ssl_state}")
+                end
+              end
               @socket.post_connection_check(@dest)
               @ssl_peer_cert = @socket.peer_cert
             end
@@ -808,7 +836,7 @@ class HTTPClient
       parse_header
       res.http_version, res.status, res.reason = @version, @status, @reason
       @headers.each do |key, value|
-        res.header.set(key, value)
+        res.header.set(key.to_s, value)
       end
       commands = @client.request_filter.collect { |filter|
         filter.filter_response(req, res)
@@ -847,17 +875,17 @@ class HTTPClient
       timeout(@receive_timeout, ReceiveTimeoutError) do
         initial_line = nil
         begin
-          initial_line = @socket.gets("\n")
-          if initial_line.nil?
+          begin
+            initial_line = @socket.gets("\n")
+            if initial_line.nil?
+              close
+              raise KeepAliveDisconnected.new(self)
+            end
+          rescue Errno::ECONNABORTED, Errno::ECONNRESET, Errno::EPIPE, IOError
+            # JRuby can raise IOError instead of ECONNRESET for now
             close
             raise KeepAliveDisconnected.new(self)
           end
-        rescue Errno::ECONNABORTED, Errno::ECONNRESET, Errno::EPIPE, IOError
-          # JRuby can raise IOError instead of ECONNRESET for now
-          close
-          raise KeepAliveDisconnected.new(self)
-        end
-        begin
           if StatusParseRegexp !~ initial_line
             @version = '0.9'
             @status = nil
