@@ -4,6 +4,7 @@ require File.expand_path('helper', File.dirname(__FILE__))
 
 class TestHTTPClient < Test::Unit::TestCase
   include Helper
+  include HTTPClient::Util
 
   def setup
     super
@@ -20,7 +21,7 @@ class TestHTTPClient < Test::Unit::TestCase
     escape_noproxy do
       @proxyio.string = ""
       @client = HTTPClient.new(proxyurl)
-      assert_equal(URI.parse(proxyurl), @client.proxy)
+      assert_equal(urify(proxyurl), @client.proxy)
       assert_equal(200, @client.head(serverurl).status)
       assert(/accept/ =~ @proxyio.string)
     end
@@ -43,7 +44,7 @@ class TestHTTPClient < Test::Unit::TestCase
     @client.get(serverurl)
     lines = str.split(/(?:\r?\n)+/)
     assert_equal("= Request", lines[0])
-    assert_match(/^From: from_bar/, lines[4])
+    assert_match(/^From: from_bar/, lines[5])
   end
 
   def test_debug_dev
@@ -76,10 +77,10 @@ class TestHTTPClient < Test::Unit::TestCase
     assert_equal("= Request", lines[0])
     assert_equal("! CONNECTION ESTABLISHED", lines[2])
     assert_equal("GET /hello HTTP/0.9", lines[3])
-    assert_equal("Connection: close", lines[5])
-    assert_equal("= Response", lines[6])
-    assert_match(/^hello$/, lines[7])
-    assert_match(/^world$/, lines[8])
+    assert_equal("Connection: close", lines[7])
+    assert_equal("= Response", lines[8])
+    assert_match(/^hello$/, lines[9])
+    assert_match(/^world$/, lines[10])
   end
 
   def test_protocol_version_http10
@@ -93,8 +94,24 @@ class TestHTTPClient < Test::Unit::TestCase
     assert_equal("= Request", lines[0])
     assert_equal("! CONNECTION ESTABLISHED", lines[2])
     assert_equal("GET /hello HTTP/1.0", lines[3])
-    assert_equal("Connection: close", lines[5])
-    assert_equal("= Response", lines[6])
+    assert_equal("Connection: close", lines[7])
+    assert_equal("= Response", lines[8])
+  end
+
+  def test_header_accept_by_default
+    str = ""
+    @client.debug_dev = str
+    @client.get(serverurl)
+    lines = str.split(/(?:\r?\n)+/)
+    assert_equal("Accept: */*", lines[5])
+  end
+
+  def test_header_accept
+    str = ""
+    @client.debug_dev = str
+    @client.get(serverurl, :header => {:Accept => 'text/html'})
+    lines = str.split(/(?:\r?\n)+/)
+    assert_equal("Accept: text/html", lines[4])
   end
 
   def test_host_given
@@ -105,7 +122,7 @@ class TestHTTPClient < Test::Unit::TestCase
     assert_equal("= Request", lines[0])
     assert_equal("! CONNECTION ESTABLISHED", lines[2])
     assert_equal("GET / HTTP/1.1", lines[3])
-    assert_equal("Host: localhost:#{serverport}", lines[5])
+    assert_equal("Host: localhost:#{serverport}", lines[7])
     #
     @client.reset_all
     str = ""
@@ -118,6 +135,14 @@ class TestHTTPClient < Test::Unit::TestCase
     assert_equal("Host: foo", lines[4]) # use given param
   end
 
+  def test_redirect_returns_not_modified
+    assert_nothing_raised do
+      timeout(2) do
+        @client.get(serverurl + 'status', {:status => 306}, {:follow_redirect => true})
+      end
+    end
+  end
+
   def test_protocol_version_http11
     assert_equal(nil, @client.protocol_version)
     str = ""
@@ -127,7 +152,7 @@ class TestHTTPClient < Test::Unit::TestCase
     assert_equal("= Request", lines[0])
     assert_equal("! CONNECTION ESTABLISHED", lines[2])
     assert_equal("GET / HTTP/1.1", lines[3])
-    assert_equal("Host: localhost:#{serverport}", lines[5])
+    assert_equal("Host: localhost:#{serverport}", lines[7])
     @client.protocol_version = 'HTTP/1.1'
     assert_equal('HTTP/1.1', @client.protocol_version)
     str = ""
@@ -150,14 +175,16 @@ class TestHTTPClient < Test::Unit::TestCase
   def test_proxy
     setup_proxyserver
     escape_noproxy do
-      assert_raises(URI::InvalidURIError) do
-       	@client.proxy = "http://"
+      begin
+        @client.proxy = "http://"
+      rescue
+        assert_match(/InvalidURIError/, $!.class.to_s)
       end
       @client.proxy = ""
       assert_nil(@client.proxy)
       @client.proxy = "http://admin:admin@foo:1234"
-      assert_equal(URI.parse("http://admin:admin@foo:1234"), @client.proxy)
-      uri = URI.parse("http://bar:2345")
+      assert_equal(urify("http://admin:admin@foo:1234"), @client.proxy)
+      uri = urify("http://bar:2345")
       @client.proxy = uri
       assert_equal(uri, @client.proxy)
       #
@@ -196,7 +223,7 @@ class TestHTTPClient < Test::Unit::TestCase
       ENV['http_proxy'] = "http://admin:admin@foo:1234"
       ENV['NO_PROXY'] = "foobar"
       client = HTTPClient.new
-      assert_equal(URI.parse("http://admin:admin@foo:1234"), client.proxy)
+      assert_equal(urify("http://admin:admin@foo:1234"), client.proxy)
       assert_equal('foobar', client.no_proxy)
     end
   end
@@ -211,7 +238,7 @@ class TestHTTPClient < Test::Unit::TestCase
       assert_equal(nil, client.proxy)
       ENV['CGI_HTTP_PROXY'] = "http://admin:admin@foo:1234"
       client = HTTPClient.new
-      assert_equal(URI.parse("http://admin:admin@foo:1234"), client.proxy)
+      assert_equal(urify("http://admin:admin@foo:1234"), client.proxy)
     end
   end
 
@@ -274,6 +301,36 @@ class TestHTTPClient < Test::Unit::TestCase
     end
   end
 
+  def test_no_proxy_with_initial_dot
+    @client.debug_dev = str = ""
+    @client.test_loopback_http_response << "HTTP/1.0 200 OK\r\n\r\n"
+    @client.no_proxy = ''
+    @client.proxy = proxyurl
+    @client.head('http://www.foo.com')
+    assert(/CONNECT TO localhost/ =~ str, 'via proxy')
+    #
+    @client.debug_dev = str = ""
+    @client.test_loopback_http_response << "HTTP/1.0 200 OK\r\n\r\n"
+    @client.no_proxy = '.foo.com'
+    @client.proxy = proxyurl
+    @client.head('http://www.foo.com')
+    assert(/CONNECT TO www.foo.com/ =~ str, 'no proxy because .foo.com matches with www.foo.com')
+    #
+    @client.debug_dev = str = ""
+    @client.test_loopback_http_response << "HTTP/1.0 200 OK\r\n\r\n"
+    @client.no_proxy = '.foo.com'
+    @client.proxy = proxyurl
+    @client.head('http://foo.com')
+    assert(/CONNECT TO localhost/ =~ str, 'via proxy because .foo.com does not matche with foo.com')
+    #
+    @client.debug_dev = str = ""
+    @client.test_loopback_http_response << "HTTP/1.0 200 OK\r\n\r\n"
+    @client.no_proxy = 'foo.com'
+    @client.proxy = proxyurl
+    @client.head('http://foo.com')
+    assert(/CONNECT TO foo.com/ =~ str, 'no proxy because foo.com matches with foo.com')
+  end
+
   def test_cookie_update_while_authentication
     escape_noproxy do
       @client.test_loopback_http_response << <<EOS
@@ -371,9 +428,20 @@ EOS
     assert_equal('message body 1', res.content)
   end
 
+  def test_request_uri_in_response
+    @client.test_loopback_http_response << "HTTP/1.0 200 OK\ncontent-length: 100\n\nmessage body"
+    assert_equal(urify('http://google.com/'), @client.get('http://google.com/').header.request_uri)
+  end
+
+  def test_request_uri_in_response_when_redirect
+    expected = urify(serverurl + 'hello')
+    assert_equal(expected, @client.get(serverurl + 'redirect1', :follow_redirect => true).header.request_uri)
+    assert_equal(expected, @client.get(serverurl + 'redirect2', :follow_redirect => true).header.request_uri)
+  end
+
   def test_redirect_non_https
     url = serverurl + 'redirect1'
-    https_url = URI.parse(url)
+    https_url = urify(url)
     https_url.scheme = 'https'
     #
     redirect_to_http = "HTTP/1.0 302 OK\nLocation: #{url}\n\n"
@@ -414,9 +482,15 @@ EOS
     end
   end
 
+  def test_redirect_see_other
+    assert_equal('hello', @client.post_content(serverurl + 'redirect_see_other'))
+  end
+
   def test_redirect_relative
     @client.test_loopback_http_response << "HTTP/1.0 302 OK\nLocation: hello\n\n"
-    assert_equal('hello', @client.get_content(serverurl + 'redirect1'))
+    silent do
+      assert_equal('hello', @client.get_content(serverurl + 'redirect1'))
+    end
     #
     @client.reset_all
     @client.redirect_uri_callback = @client.method(:strict_redirect_uri_callback)
@@ -433,11 +507,13 @@ EOS
 
   def test_redirect_https_relative
     url = serverurl + 'redirect1'
-    https_url = URI.parse(url)
+    https_url = urify(url)
     https_url.scheme = 'https'
     @client.test_loopback_http_response << "HTTP/1.0 302 OK\nLocation: /foo\n\n"
     @client.test_loopback_http_response << "HTTP/1.0 200 OK\n\nhello"
-    assert_equal('hello', @client.get_content(https_url))
+    silent do
+      assert_equal('hello', @client.get_content(https_url))
+    end
   end
 
   def test_no_content
@@ -538,29 +614,39 @@ EOS
 
   def test_head
     assert_equal("head", @client.head(serverurl + 'servlet').header["x-head"][0])
-    res = @client.head(serverurl + 'servlet', {1=>2, 3=>4})
-    assert_equal('1=2&3=4', res.header["x-query"][0])
+    param = {'1'=>'2', '3'=>'4'}
+    res = @client.head(serverurl + 'servlet', param)
+    assert_equal(param, params(res.header["x-query"][0]))
   end
 
   def test_head_async
-    conn = @client.head_async(serverurl + 'servlet', {1=>2, 3=>4})
+    param = {'1'=>'2', '3'=>'4'}
+    conn = @client.head_async(serverurl + 'servlet', param)
     Thread.pass while !conn.finished?
     res = conn.pop
-    assert_equal('1=2&3=4', res.header["x-query"][0])
+    assert_equal(param, params(res.header["x-query"][0]))
   end
 
   def test_get
     assert_equal("get", @client.get(serverurl + 'servlet').content)
-    res = @client.get(serverurl + 'servlet', {1=>2, 3=>4})
-    assert_equal('1=2&3=4', res.header["x-query"][0])
+    param = {'1'=>'2', '3'=>'4'}
+    res = @client.get(serverurl + 'servlet', param)
+    assert_equal(param, params(res.header["x-query"][0]))
     assert_nil(res.contenttype)
     #
     url = serverurl.to_s + 'servlet?5=6&7=8'
-    res = @client.get(url, {1=>2, 3=>4})
-    assert_equal('1=2&3=4&5=6&7=8', res.header["x-query"][0])
+    res = @client.get(url, param)
+    assert_equal(param.merge("5"=>"6", "7"=>"8"), params(res.header["x-query"][0]))
     assert_nil(res.contenttype)
   end
 
+  def test_head_follow_redirect
+    expected = urify(serverurl + 'hello')
+    assert_equal(expected, @client.head(serverurl + 'hello', :follow_redirect => true).header.request_uri)
+    assert_equal(expected, @client.head(serverurl + 'redirect1', :follow_redirect => true).header.request_uri)
+    assert_equal(expected, @client.head(serverurl + 'redirect2', :follow_redirect => true).header.request_uri)
+  end
+
   def test_get_follow_redirect
     assert_equal('hello', @client.get(serverurl + 'hello', :follow_redirect => true).body)
     assert_equal('hello', @client.get(serverurl + 'redirect1', :follow_redirect => true).body)
@@ -568,10 +654,11 @@ EOS
   end
 
   def test_get_async
-    conn = @client.get_async(serverurl + 'servlet', {1=>2, 3=>4})
+    param = {'1'=>'2', '3'=>'4'}
+    conn = @client.get_async(serverurl + 'servlet', param)
     Thread.pass while !conn.finished?
     res = conn.pop
-    assert_equal('1=2&3=4', res.header["x-query"][0])
+    assert_equal(param, params(res.header["x-query"][0]))
   end
 
   def test_get_async_for_largebody
@@ -603,8 +690,9 @@ EOS
 
   def test_post
     assert_equal("post", @client.post(serverurl + 'servlet').content[0, 4])
-    res = @client.post(serverurl + 'servlet', {1=>2, 3=>4})
-    assert_equal('1=2&3=4', res.header["x-query"][0])
+    param = {'1'=>'2', '3'=>'4'}
+    res = @client.post(serverurl + 'servlet', param)
+    assert_equal(param, params(res.header["x-query"][0]))
   end
 
   def test_post_follow_redirect
@@ -614,25 +702,42 @@ EOS
   end
 
   def test_post_with_content_type
+    param = [['1', '2'], ['3', '4']]
     ext = {'content-type' => 'application/x-www-form-urlencoded', 'hello' => 'world'}
     assert_equal("post", @client.post(serverurl + 'servlet').content[0, 4], ext)
-    res = @client.post(serverurl + 'servlet', {1=>2, 3=>4}, ext)
-    assert_equal('1=2&3=4', res.header["x-query"][0])
+    res = @client.post(serverurl + 'servlet', param, ext)
+    assert_equal(Hash[param], params(res.header["x-query"][0]))
     #
     ext = [['content-type', 'multipart/form-data'], ['hello', 'world']]
     assert_equal("post", @client.post(serverurl + 'servlet').content[0, 4], ext)
-    res = @client.post(serverurl + 'servlet', {1=>2, 3=>4}, ext)
+    res = @client.post(serverurl + 'servlet', param, ext)
     assert_match(/Content-Disposition: form-data; name="1"/, res.content)
     assert_match(/Content-Disposition: form-data; name="3"/, res.content)
     #
     ext = {'content-type' => 'multipart/form-data; boundary=hello'}
     assert_equal("post", @client.post(serverurl + 'servlet').content[0, 4], ext)
-    res = @client.post(serverurl + 'servlet', {1=>2, 3=>4}, ext)
+    res = @client.post(serverurl + 'servlet', param, ext)
     assert_match(/Content-Disposition: form-data; name="1"/, res.content)
     assert_match(/Content-Disposition: form-data; name="3"/, res.content)
     assert_equal("post,--hello\r\nContent-Disposition: form-data; name=\"1\"\r\n\r\n2\r\n--hello\r\nContent-Disposition: form-data; name=\"3\"\r\n\r\n4\r\n--hello--\r\n\r\n", res.content)
   end
 
+  def test_post_with_custom_multipart_and_boolean_params
+    param = [['boolean_true', true]]
+    ext = { 'content-type' => 'multipart/form-data' }
+    assert_equal("post", @client.post(serverurl + 'servlet').content[0, 4], ext)
+    res = @client.post(serverurl + 'servlet', param, ext)
+    assert_match(/Content-Disposition: form-data; name="boolean_true"\r\n\r\ntrue\r\n/, res.content)
+    #
+    param = [['boolean_false', false]]
+    res = @client.post(serverurl + 'servlet', param, ext)
+    assert_match(/Content-Disposition: form-data; name="boolean_false"\r\n\r\nfalse\r\n/, res.content)
+    #
+    param = [['nil', nil]]
+    res = @client.post(serverurl + 'servlet', param, ext)
+    assert_match(/Content-Disposition: form-data; name="nil"\r\n\r\n\r\n/, res.content)
+  end
+
   def test_post_with_file
     STDOUT.sync = true
     File.open(__FILE__) do |file|
@@ -643,6 +748,19 @@ EOS
     end
   end
 
+  def test_post_with_file_without_size
+    STDOUT.sync = true
+    File.open(__FILE__) do |file|
+      def file.size
+        # Simulates some strange Windows behaviour
+        raise SystemCallError.new "Unknown Error (20047)"
+      end
+      assert_nothing_raised do
+        @client.post(serverurl + 'servlet', {1=>2, 3=>file})
+      end
+    end
+  end
+
   def test_post_with_io # streaming, but not chunked
     myio = StringIO.new("X" * (HTTP::Message::Body::DEFAULT_CHUNK_SIZE + 1))
     def myio.read(*args)
@@ -677,10 +795,11 @@ EOS
   end
 
   def test_post_async
-    conn = @client.post_async(serverurl + 'servlet', {1=>2, 3=>4})
+    param = {'1'=>'2', '3'=>'4'}
+    conn = @client.post_async(serverurl + 'servlet', param)
     Thread.pass while !conn.finished?
     res = conn.pop
-    assert_equal('1=2&3=4', res.header["x-query"][0])
+    assert_equal(param, params(res.header["x-query"][0]))
   end
 
   def test_post_with_block
@@ -693,7 +812,8 @@ EOS
     assert_nil(res.content)
     #
     called = false
-    res = @client.post(serverurl + 'servlet', {1=>2, 3=>4}) { |str|
+    param = [['1', '2'], ['3', '4']]
+    res = @client.post(serverurl + 'servlet', param) { |str|
       assert_equal('post,1=2&3=4', str)
       called = true
     }
@@ -735,8 +855,11 @@ EOS
 
   def test_put
     assert_equal("put", @client.put(serverurl + 'servlet').content)
-    res = @client.put(serverurl + 'servlet', {1=>2, 3=>4})
-    assert_equal('1=2&3=4', res.header["x-query"][0])
+    param = {'1'=>'2', '3'=>'4'}
+    @client.debug_dev = str = ''
+    res = @client.put(serverurl + 'servlet', param)
+    assert_equal(param, params(res.header["x-query"][0]))
+    assert_equal('Content-Type: application/x-www-form-urlencoded', str.split(/\r?\n/)[5])
   end
 
   def test_put_bytesize
@@ -746,16 +869,25 @@ EOS
   end
 
   def test_put_async
-    conn = @client.put_async(serverurl + 'servlet', {1=>2, 3=>4})
+    param = {'1'=>'2', '3'=>'4'}
+    conn = @client.put_async(serverurl + 'servlet', param)
     Thread.pass while !conn.finished?
     res = conn.pop
-    assert_equal('1=2&3=4', res.header["x-query"][0])
+    assert_equal(param, params(res.header["x-query"][0]))
   end
 
   def test_delete
     assert_equal("delete", @client.delete(serverurl + 'servlet').content)
   end
 
+  # Not prohibited by spec, but normally it's ignored
+  def test_delete_with_body
+    param = {'1'=>'2', '3'=>'4'}
+    @client.debug_dev = str = ''
+    assert_equal("delete", @client.delete(serverurl + 'servlet', param).content)
+    assert_equal({'1' => ['2'], '3' => ['4']}, HTTP::Message.parse(str.split(/\r?\n\r?\n/)[2]))
+  end
+
   def test_delete_async
     conn = @client.delete_async(serverurl + 'servlet')
     Thread.pass while !conn.finished?
@@ -787,30 +919,34 @@ EOS
 
   def test_proppatch
     assert_equal("proppatch", @client.proppatch(serverurl + 'servlet').content)
-    res = @client.proppatch(serverurl + 'servlet', {1=>2, 3=>4})
+    param = {'1'=>'2', '3'=>'4'}
+    res = @client.proppatch(serverurl + 'servlet', param)
     assert_equal('proppatch', res.content)
-    assert_equal('1=2&3=4', res.header["x-query"][0])
+    assert_equal(param, params(res.header["x-query"][0]))
   end
 
   def test_proppatch_async
-    conn = @client.proppatch_async(serverurl + 'servlet', {1=>2, 3=>4})
+    param = {'1'=>'2', '3'=>'4'}
+    conn = @client.proppatch_async(serverurl + 'servlet', param)
     Thread.pass while !conn.finished?
     res = conn.pop
     assert_equal('proppatch', res.content.read)
-    assert_equal('1=2&3=4', res.header["x-query"][0])
+    assert_equal(param, params(res.header["x-query"][0]))
   end
 
   def test_trace
     assert_equal("trace", @client.trace(serverurl + 'servlet').content)
-    res = @client.trace(serverurl + 'servlet', {1=>2, 3=>4})
-    assert_equal('1=2&3=4', res.header["x-query"][0])
+    param = {'1'=>'2', '3'=>'4'}
+    res = @client.trace(serverurl + 'servlet', param)
+    assert_equal(param, params(res.header["x-query"][0]))
   end
 
   def test_trace_async
-    conn = @client.trace_async(serverurl + 'servlet', {1=>2, 3=>4})
+    param = {'1'=>'2', '3'=>'4'}
+    conn = @client.trace_async(serverurl + 'servlet', param)
     Thread.pass while !conn.finished?
     res = conn.pop
-    assert_equal('1=2&3=4', res.header["x-query"][0])
+    assert_equal(param, params(res.header["x-query"][0]))
   end
 
   def test_chunked
@@ -835,6 +971,18 @@ EOS
     assert_equal({}, check_query_get(''))
     assert_equal({'1'=>'2'}, check_query_get({1=>StringIO.new('2')}))
     assert_equal({'1'=>'2', '3'=>'4'}, check_query_get(StringIO.new('3=4&1=2')))
+
+    hash = check_query_get({"a"=>["A","a"], "B"=>"b"})
+    assert_equal({'a'=>'A', 'B'=>'b'}, hash)
+    assert_equal(['A','a'], hash['a'].to_ary)
+
+    hash = check_query_get({"a"=>WEBrick::HTTPUtils::FormData.new("A","a"), "B"=>"b"})
+    assert_equal({'a'=>'A', 'B'=>'b'}, hash)
+    assert_equal(['A','a'], hash['a'].to_ary)
+
+    hash = check_query_get({"a"=>[StringIO.new("A"),StringIO.new("a")], "B"=>StringIO.new("b")})
+    assert_equal({'a'=>'A', 'B'=>'b'}, hash)
+    assert_equal(['A','a'], hash['a'].to_ary)
   end
 
   def test_post_body
@@ -874,6 +1022,13 @@ EOS
     assert_match("ABC: DEF", lines[5])
   end
 
+  def test_http_custom_date_header
+    @client.debug_dev = (str = "")
+    res = @client.get(serverurl + 'hello', :header => {'Date' => 'foo'})
+    lines = str.split(/(?:\r?\n)+/)
+    assert_equal('Date: foo', lines[4])
+  end
+
   def test_timeout
     assert_equal(60, @client.connect_timeout)
     assert_equal(120, @client.send_timeout)
@@ -948,7 +1103,7 @@ EOS
   def test_cookies
     cookiefile = File.join(File.dirname(File.expand_path(__FILE__)), 'test_cookies_file')
     File.open(cookiefile, "wb") do |f|
-      f << "http://rubyforge.org/account/login.php	session_ser	LjEwMy45Ni40Ni0q%2A-fa0537de8cc31	2000000000	.rubyforge.org	/	13\n"
+      f << "http://rubyforge.org/account/login.php\tsession_ser\tLjEwMy45Ni40Ni0q%2A-fa0537de8cc31\t2000000000\t.rubyforge.org\t/\t13\n"
     end
     @client.set_cookie_store(cookiefile)
     cookie = @client.cookie_manager.cookies.first
@@ -960,7 +1115,7 @@ EOS
     @client.get_content('http://rubyforge.org/account/login.php')
     @client.save_cookie_store
     str = File.read(cookiefile)
-    assert_match(%r(http://rubyforge.org/account/login.php	foo	bar	1924873200	rubyforge.org	/account	1), str)
+    assert_match(%r(http://rubyforge.org/account/login.php\tfoo\tbar\t1924873200\trubyforge.org\t/account\t1), str)
     File.unlink(cookiefile)
   end
 
@@ -1024,8 +1179,8 @@ EOS
     extend HTTPClient::Util
     assert_nil(urify(nil))
     uri = 'http://foo'
-    assert_equal(URI.parse(uri), urify(uri))
-    assert_equal(URI.parse(uri), urify(URI.parse(uri)))
+    assert_equal(urify(uri), urify(uri))
+    assert_equal(urify(uri), urify(urify(uri)))
   end
 
   def test_connection
@@ -1045,7 +1200,7 @@ EOS
       site.inspect
     end
     #
-    site = HTTPClient::Site.new(URI.parse('http://localhost:12345/foo'))
+    site = HTTPClient::Site.new(urify('http://localhost:12345/foo'))
     assert_equal('http', site.scheme)
     assert_equal('localhost', site.host)
     assert_equal(12345, site.port)
@@ -1055,9 +1210,9 @@ EOS
       site.inspect
     end
     #
-    site1 = HTTPClient::Site.new(URI.parse('http://localhost:12341/'))
-    site2 = HTTPClient::Site.new(URI.parse('http://localhost:12342/'))
-    site3 = HTTPClient::Site.new(URI.parse('http://localhost:12342/'))
+    site1 = HTTPClient::Site.new(urify('http://localhost:12341/'))
+    site2 = HTTPClient::Site.new(urify('http://localhost:12342/'))
+    site3 = HTTPClient::Site.new(urify('http://localhost:12342/'))
     assert(!(site1 == site2))
     h = { site1 => 'site1', site2 => 'site2' }
     h[site3] = 'site3'
@@ -1110,9 +1265,9 @@ EOS
   end
 
   def test_connect_request
-    req = HTTP::Message.new_connect_request(URI.parse('https://foo/bar'))
+    req = HTTP::Message.new_connect_request(urify('https://foo/bar'))
     assert_equal("CONNECT foo:443 HTTP/1.0\r\n\r\n", req.dump)
-    req = HTTP::Message.new_connect_request(URI.parse('https://example.com/'))
+    req = HTTP::Message.new_connect_request(urify('https://example.com/'))
     assert_equal("CONNECT example.com:443 HTTP/1.0\r\n\r\n", req.dump)
   end
 
@@ -1211,6 +1366,17 @@ EOS
     assert_equal('PART_NUMBER', res.cookies[1].name)
   end
 
+  def test_ok_response_success
+    res = HTTP::Message.new_response('response')
+    assert_equal(true, res.ok?)
+    res.status = 404
+    assert_equal(false, res.ok?)
+    res.status = 500
+    assert_equal(false, res.ok?)
+    res.status = 302
+    assert_equal(false, res.ok?)
+  end
+
   if !defined?(JRUBY_VERSION) and RUBY_VERSION < '1.9'
     def test_timeout_scheduler
       assert_equal('hello', @client.get_content(serverurl + 'hello'))
@@ -1357,12 +1523,14 @@ EOS
   end
 
   def test_socket_local
-    @client.socket_local.host = 'localhost'
+    @client.socket_local.host = '127.0.0.1'
     assert_equal('hello', @client.get_content(serverurl + 'hello'))
     @client.reset_all
     @client.socket_local.port = serverport
-    assert_raises(Errno::EADDRINUSE) do
-      assert_equal('hello', @client.get_content(serverurl + 'hello'))
+    begin
+      @client.get_content(serverurl + 'hello')
+    rescue Errno::EADDRINUSE, SocketError
+      assert(true)
     end
   end
 
@@ -1379,6 +1547,40 @@ EOS
     end
   end
 
+  if RUBY_VERSION >= "1.9.3"
+    def test_continue
+      @client.debug_dev = str = ''
+      res = @client.get(serverurl + 'continue', :header => {:Expect => '100-continue'})
+      assert_equal(200, res.status)
+      assert_equal('done!', res.body)
+      assert_match(/Expect: 100-continue/, str)
+    end
+  end
+
+  def test_ipv6literaladdress_in_uri
+    server = TCPServer.open('::1', 0) rescue return # Skip if IPv6 is unavailable.
+    server_thread = Thread.new {
+      Thread.abort_on_exception = true
+      sock = server.accept
+      while line = sock.gets
+        break if line.chomp.empty?
+      end
+      sock.write("HTTP/1.1 200 OK\r\n")
+      sock.write("Content-Length: 5\r\n")
+      sock.write("\r\n")
+      sock.write("12345")
+      sock.close
+    }
+    uri = "http://[::1]:#{server.addr[1]}/"
+    begin
+      assert_equal('12345', @client.get(uri).body)
+    ensure
+      server.close
+      server_thread.kill
+      server_thread.join
+    end
+  end
+
 private
 
   def check_query_get(query)
@@ -1402,25 +1604,20 @@ private
       :DocumentRoot => File.dirname(File.expand_path(__FILE__))
     )
     @serverport = @server.config[:Port]
-    [:hello, :sleep, :servlet_redirect, :redirect1, :redirect2, :redirect3, :redirect_self, :relative_redirect, :chunked, :largebody, :status, :compressed, :charset].each do |sym|
+    [
+      :hello, :sleep, :servlet_redirect, :redirect1, :redirect2, :redirect3,
+      :redirect_self, :relative_redirect, :redirect_see_other, :chunked,
+      :largebody, :status, :compressed, :charset, :continue
+    ].each do |sym|
       @server.mount(
-	"/#{sym}",
-	WEBrick::HTTPServlet::ProcHandler.new(method("do_#{sym}").to_proc)
+        "/#{sym}",
+        WEBrick::HTTPServlet::ProcHandler.new(method("do_#{sym}").to_proc)
       )
     end
     @server.mount('/servlet', TestServlet.new(@server))
     @server_thread = start_server_thread(@server)
   end
 
-  def escape_env
-    env = {}
-    env.update(ENV)
-    yield
-  ensure
-    ENV.clear
-    ENV.update(env)
-  end
-
   def escape_noproxy
     backup = HTTPClient::NO_PROXY_HOSTS.dup
     HTTPClient::NO_PROXY_HOSTS.clear
@@ -1442,11 +1639,11 @@ private
   end
 
   def do_servlet_redirect(req, res)
-    res.set_redirect(WEBrick::HTTPStatus::Found, serverurl + "servlet") 
+    res.set_redirect(WEBrick::HTTPStatus::Found, serverurl + "servlet")
   end
 
   def do_redirect1(req, res)
-    res.set_redirect(WEBrick::HTTPStatus::MovedPermanently, serverurl + "hello") 
+    res.set_redirect(WEBrick::HTTPStatus::MovedPermanently, serverurl + "hello")
   end
 
   def do_redirect2(req, res)
@@ -1454,15 +1651,23 @@ private
   end
 
   def do_redirect3(req, res)
-    res.set_redirect(WEBrick::HTTPStatus::Found, serverurl + "hello") 
+    res.set_redirect(WEBrick::HTTPStatus::Found, serverurl + "hello")
   end
 
   def do_redirect_self(req, res)
-    res.set_redirect(WEBrick::HTTPStatus::Found, serverurl + "redirect_self") 
+    res.set_redirect(WEBrick::HTTPStatus::Found, serverurl + "redirect_self")
   end
 
   def do_relative_redirect(req, res)
-    res.set_redirect(WEBrick::HTTPStatus::Found, "hello") 
+    res.set_redirect(WEBrick::HTTPStatus::Found, "hello")
+  end
+
+  def do_redirect_see_other(req, res)
+    if req.request_method == 'POST'
+      res.set_redirect(WEBrick::HTTPStatus::SeeOther, serverurl + "redirect_see_other") # self
+    else
+      res.body = 'hello'
+    end
   end
 
   def do_chunked(req, res)
@@ -1502,13 +1707,18 @@ private
     res.status = req.query['status'].to_i
   end
 
+  def do_continue(req, res)
+    req.continue
+    res.body = 'done!'
+  end
+
   class TestServlet < WEBrick::HTTPServlet::AbstractServlet
     def get_instance(*arg)
       self
     end
 
     def do_HEAD(req, res)
-      res["x-head"] = 'head'	# use this for test purpose only.
+      res["x-head"] = 'head'    # use this for test purpose only.
       res["x-query"] = query_response(req)
     end
 
@@ -1568,9 +1778,9 @@ private
     def query_escape(query)
       escaped = []
       query.sort_by { |k, v| k }.collect do |k, v|
-	v.to_ary.each do |ve|
-	  escaped << CGI.escape(k) + '=' + CGI.escape(ve)
-	end
+        v.to_ary.each do |ve|
+          escaped << CGI.escape(k) + '=' + CGI.escape(ve)
+        end
       end
       escaped.join('&')
     end
