@@ -34,7 +34,7 @@ class TestSSL < Test::Unit::TestCase
     assert_nil(cfg.verify_callback)
     assert_nil(cfg.timeout)
     assert_equal(OpenSSL::SSL::OP_ALL | OpenSSL::SSL::OP_NO_SSLv2, cfg.options)
-    assert_equal("ALL:!ADH:!LOW:!EXP:!MD5:+SSLv2:@STRENGTH", cfg.ciphers)
+    assert_equal("ALL:!aNULL:!eNULL:!SSLv2", cfg.ciphers)
     assert_instance_of(OpenSSL::X509::Store, cfg.cert_store)
   end
 
@@ -155,6 +155,17 @@ end
     assert_equal("hello", @client.get_content(@url))
   end
 
+  def test_set_default_paths
+    assert_raise(OpenSSL::SSL::SSLError) do
+      @client.get(@url)
+    end
+    escape_env do
+      ENV['SSL_CERT_FILE'] = File.join(DIR, 'ca-chain.cert')
+      @client.ssl_config.set_default_paths
+      @client.get(@url)
+    end
+  end
+
 private
 
   def cert(filename)
